/*
 * Main program.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include "common.h"
#include "config.h"
#include "program.h"
#include "templates.h"
#include "version.h"

/* Parser functions and settings. */

extern void yyparse(void);
extern void yyrestart(FILE *);
extern int yydebug;

/* Specialised parser functions. */

extern void reset(void);



/* Return a copy of the given string without any extension. */

static char *make_prefix(const char *s)
{
  char *copy = strdup(s);
  char *ext;

  if (copy == NULL)
    return NULL;

  ext = strrchr(copy, '.');

  if (ext != NULL)
    *ext = '\0';

  return copy;
}

static int get_compound_prefix_conflict(int argc, char *argv[])
{
  char *arg, *prefix, *name;
  int index, result;

  for (index = optind; index < argc; index++)
  {
    arg = argv[index];
    prefix = make_prefix(arg);
    result = 0;

    if (prefix != NULL)
    {
      name = make_basename(prefix);

      if (name != NULL)
      {
        result = strcasecmp(name, conf.compound);
        free(name);
      }

      free(prefix);
    }

    if (!result)
      return index;
  }

  return 0;
}


/* Option definitions. */

static struct option long_options[] = {
/* long opt       following           var pointer short opt */
  {"all",         no_argument,        NULL,       'a'       },
  {"client",      no_argument,        NULL,       'c'       },
  {"comp",        required_argument,  NULL,       'C'       },
  {"dir",         required_argument,  NULL,       'd'       },
  {"headers",     no_argument,        NULL,       'h'       },
  {"help",        no_argument,        NULL,       '?'       },
  {"interfaces",  no_argument,        NULL,       'i'       },
  {"language",    required_argument,  NULL,       'l'       },
  {"comp-name",   required_argument,  NULL,       'N'       },
  {"routines",    no_argument,        NULL,       'r'       },
  {"server",      no_argument,        NULL,       's'       },
  {"verbose",     no_argument,        NULL,       'v'       },
  {"version",     no_argument,        NULL,       'V'       },
  {0,             0,                  0,          0         },
  };


/* Main program, parsing indicated files or standard input. */

int main(int argc, char *argv[])
{
  FILE *fp;
  char *arg, *progname;
  int option, index, selected_content = 0, selected_role = 0;

#if YYDEBUG == 1
  yydebug = 0;
#endif

  /* Accept various options. */

  while (1)
  {
    option = getopt_long(argc, argv, "acC:d:hil:N:rsvV?", long_options, NULL);
    if (option == -1)
      break;

    switch (option)
    {
      /* Generate everything. */
      case 'a':
        conf.client = 1; conf.server = 1;
        conf.headers = 1; conf.interfaces = 1; conf.routines = 1;
        selected_content = 1; 
        selected_role = 1; 
        break;

      /* Select client output. */
      case 'c':
        conf.client = 1;
        selected_role = 1;
        break;

      /* Select compound interface filename prefix and output. */
      case 'C':
        conf.compound = optarg;
        if (conf.compound_name == NULL)
          conf.compound_name = conf.compound;
        selected_role = 1;
        break;

      /* Set output directory. */
      case 'd':
        conf.output_dir = optarg;
        break;

      /* Generate headers. */
      case 'h':
        conf.headers = 1;
        selected_content = 1;
        break;

      /* Generate interface headers. */
      case 'i':
        conf.interfaces = 1;
        selected_content = 1;
        break;

      /* Select output language. */
      case 'l':
        if (!strcasecmp(optarg, "c"))
          conf.language = C_LANGUAGE;
        else if (!strcasecmp(optarg, "c++"))
          conf.language = CPP_LANGUAGE;
        else
        {
          fprintf(stderr, "Output language not recognised: %s\n", optarg);
          return 1;
        }
        break;

      /* Select compound interface name and output. */
      case 'N':
        conf.compound_name = optarg;
        if (conf.compound == NULL)
          conf.compound = conf.compound_name;
        selected_role = 1;
        break;

      /* Generate routines (definitions). */
      case 'r':
        conf.routines = 1;
        selected_content = 1;
        break;

      /* Select server output. */
      case 's':
        conf.server = 1;
        selected_role = 1;
        break;

      /* Verbose reporting. */
      case 'v':
        conf.verbose = 1;
        break;

      /* Show version (defined in the Makefile) and exit. */
      case 'V':
        puts(VERSION_INFORMATION);
        return 0;

      /* Show help and exit. */
      case '?':
        progname = make_basename(argv[0]);
        printf(help_text, progname);
        free(progname);
        return 0;

      default:
        return 1;
    }
  }

  /* Without any output selection, enable all output of the given kind. */

  if (!selected_content)
  {
    conf.headers = 1; conf.interfaces = 1; conf.routines = 1;
  }

  if (!selected_role)
  {
    conf.client = 1; conf.server = 1;
  }

  /* Check any compound interface prefix against named files. */

  if (conf.compound != NULL)
  {
    if ((index = get_compound_prefix_conflict(argc, argv)))
    {
      fprintf(stderr, "Compound interface prefix %s conflicts with filename %s.\n",
                      conf.compound, argv[index]);
      return 1;
    }
  }

  /* Produce an error without any input files. */

  if (optind >= argc)
  {
    fputs("No input files.\n", stderr);
    return 1;
  }

  /* Begin generating any compound interface code. */

  if (!begin_compound_output())
  {
    end_compound_output();
    return 1;
  }

  /* Process named files. */

  for (index = optind; index < argc; index++)
  {
    arg = argv[index];

    /* Handle actual files. */

    conf.output_prefix = make_prefix(arg);

    if (conf.output_prefix != NULL)
    {
      fp = fopen(arg, "r");

      if (fp != NULL)
      {
        yyrestart(fp);
        yyparse();
        fclose(fp);

        /* Reset any specialised parser state. */

        reset();
      }

      free(conf.output_prefix);

      if (fp == NULL)
      {
        fprintf(stderr, "Could not open file: %s\n", arg);
        return 1;
      }
    }
  }

  /* End generating any compound interface code. */

  end_compound_output();

  return 0;
}
